@@ -3,7 +3,7 @@
  dvd-vr.c     Identify and optionally copy the individual programs
               from a DVD-VR format disc
 
- Copyright © 2007-2009 Pádraig Brady <P@draigBrady.com>
+ Copyright © 2007-2010 Pádraig Brady <P@draigBrady.com>
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
@@ -29,7 +29,8 @@ Notes:
     Merged programs are not handled yet though as
     I would need to fully parse the higher level program set info.
     Note the VOBs output from this program can be trivially
-    concatenated with the unix cat command for example.
+    concatenated with the unix cat command for example
+    (note there will be timestamp jumps which may be problematic).
 
     While extracting the DVD data, this program instructs the system
     to not cache the data so that existing cached data is not affected.
@@ -56,7 +57,7 @@ Requirements:
     Tested on linux, CYGWIN and Mac OS X
 */
 
-#define _GNU_SOURCE           /* for posix_fadvise() and futimes() */
+#define _GNU_SOURCE           /* for posix_fadvise(), futimes() */
 #define _FILE_OFFSET_BITS 64  /* for implicit large file support */
 
 #include <inttypes.h>
@@ -65,6 +66,7 @@ Requirements:
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <wchar.h>
 #include <unistd.h>
 #include <fcntl.h>
 #include <time.h>
@@ -78,6 +80,14 @@ Requirements:
 #include <errno.h>
 #include <limits.h>
 
+#if !defined(MB_LEN_MAX) || MB_LEN_MAX<16
+/* 1 char could be converted to 2 multibyte chars
+ * (for example combining accents), with each taking up to
+ * 6 bytes in UTF-8 for example */
+# undef MB_LEN_MAX
+# define MB_LEN_MAX 16
+#endif
+
 #define TYPE_SIGNED(t) (! ((t) 0 < (t) -1))
 #define TYPE_MAX(t) \
   ((t) (! TYPE_SIGNED (t) \
@@ -108,8 +118,8 @@ const char* base_name = TIMESTAMP_FMT;
 
 FILE* stdinfo; /* Where we write disc info */
 
-#ifdef HAVE_ICONV
 #include <langinfo.h>
+#ifdef HAVE_ICONV
 #include <iconv.h>
 #endif
 const char* disc_charset;
@@ -118,6 +128,24 @@ const char* sys_charset;
 /*********************************************************************************
  *                          support routines
  *********************************************************************************/
+static size_t my_strnlen(const char* s, size_t n)
+{
+    size_t len = 0;
+    while (n-- && *s++) len++;
+    return len;
+}
+
+/* Mac OS X doesn't provide strndup :( */
+static char* my_strndup(const char *s, size_t n)
+{
+    size_t len = my_strnlen(s, n);
+    char* ret = malloc(len+1);
+    if (ret) {
+        memcpy(ret, s, len);
+        ret[len] = '\0';
+    }
+    return ret;
+}
 
 #ifndef NDEBUG
 void hexdump(const void* data, int len)
@@ -130,6 +158,7 @@ void hexdump(const void* data, int len)
     }
     if (len%16) putchar('\n');
 }
+
 #endif//NDEBUG
 
 typedef enum {
@@ -191,33 +220,32 @@ typedef void (*process_func_t)(uint8_t* buf, unsigned int bs, void* context);
   Copy data between file descriptors while not
   putting more than blocks*block_size in the system cache.
   Therefore you will probably want to call this function repeatedly.
+
+  I tested 3 methods for streaming large amounts of data to/from disk.
+  All 3 took the same time as the bottleneck is the reading and writing to disk.
+  On x86 at least there is no significant difference between the AUTO and ALLOC_ALIGN
+  methods, the latter of which allocates the userspace buffer aligned on a page.
+  There was a noticeable reduction in CPU usage when MMAP_WRITE was used,
+  but the CPU usage is insignificant anyway due to the disc speeds we will
+  generally be dealing with. I also noticed that the MMAP method was more stable
+  giving consistent timings in all benchmark runs. However to ease portability worries
+  I use the AUTO method below, which will also allow us to modify the MPEG frames if
+  required. For reference the timings for extracting a 338 MiB VOB
+  from a VRO on the same hard disk were:
+
+      MMAP_WRITE
+        real    0m30.650s
+        user    0m0.007s
+        sys     0m1.130s
+      AUTO/ALLOC_ALIGN
+        real    0m31.776s
+        user    0m0.075s
+        sys     0m1.803s
  */
+
 static int stream_data(int src_fd, int dst_fd, uint32_t blocks, uint16_t block_size,
                        process_func_t process_func, void* process_context)
 {
-
-/* I tested 3 methods for streaming large amounts of data to/from disk.
- * All 3 took the same time as the bottleneck is the reading and writing to disk.
- * On x86 at least there is no significant difference between the AUTO and ALLOC_ALIGN
- * methods, the latter of which allocates the userspace buffer aligned on a page.
- * There was a noticeable reduction in CPU usage when MMAP_WRITE was used,
- * but the CPU usage is insignificant anyway due to the disc speeds we will
- * generally be dealing with. I also noticed that the MMAP method was more stable
- * giving consistent timings in all benchmark runs. However to ease portability worries
- * I use the AUTO method below, which will also allow us to modify the MPEG frames if
- * required. For reference the timings for extracting a 338 MiB VOB
- * from a VRO on the same hard disk were:
- *
- *     MMAP_WRITE
- *       real    0m30.650s
- *       user    0m0.007s
- *       sys     0m1.130s
- *     AUTO/ALLOC_ALIGN
- *       real    0m31.776s
- *       user    0m0.075s
- *       sys     0m1.803s
- */
-
 #define AUTO
 #define BLOCKS_PER_OP 1
 
@@ -276,18 +304,25 @@ static int stream_data(int src_fd, int dst_fd, uint32_t blocks, uint16_t block_s
     so that we don't dump any readahead cache. */
     uint32_t bytes = blocks * block_size;
     off_t offset = lseek(src_fd, 0, SEEK_CUR);
-    int ret = posix_fadvise(src_fd, offset-bytes, bytes, POSIX_FADV_DONTNEED);
-    if (ret) {
-        fprintf(stderr, "Warning: posix_fadvise failed [%s]\n", strerror(ret));
+    /* Note src is already guaranteed seekable, but offset may
+     * be 0 for example if /dev/zero is specified for testing. */
+    if (offset >= bytes) {
+        int ret = posix_fadvise(src_fd, offset-bytes, bytes, POSIX_FADV_DONTNEED);
+        if (ret) {
+            fprintf(stderr, "Warning: posix_fadvise failed [%s]\n", strerror(ret));
+        }
     }
 
     /* Don't fill cache with DST.
     Note this slows the operation down by 20% when both source
     and dest are on the same hard disk at least. I guess
     this is due to implicit syncing in posix_fadvise()? */
-    ret = posix_fadvise(dst_fd, 0, 0, POSIX_FADV_DONTNEED);
-    if (ret) {
-        fprintf(stderr, "Warning: posix_fadvise failed [%s]\n", strerror(ret));
+    offset = lseek(dst_fd, 0, SEEK_CUR);
+    if (offset != (off_t)-1) { /* seekable */
+        int ret = posix_fadvise(dst_fd, 0, 0, POSIX_FADV_DONTNEED);
+        if (ret) {
+            fprintf(stderr, "Warning: posix_fadvise failed [%s]\n", strerror(ret));
+        }
     }
 #endif //POSIX_FADV_DONTNEED
 
@@ -391,9 +426,7 @@ static bool text_convert(const char *src, size_t srclen, char *dst, size_t dstle
 #else
     /* avoid warnings (__attribute__ ((unused)) is too verbose/non standard) */
     (void)src; (void)dst; (void)srclen; (void)dstlen;
-    fprintf(stderr,
-            "Error converting text. libiconv missing\n",
-            disc_charset);
+    fprintf(stderr, "Error converting text. libiconv missing\n");
 #endif
     return ret;
 }
@@ -409,8 +442,16 @@ typedef struct {
 } p_video_attr_t;
 p_video_attr_t* ifo_video_attrs;
 
+typedef enum {
+    SCRAMBLED_UNSET=-1,
+    UNSCRAMBLED=0,
+    SCRAMBLED=1,
+    PARTIALLY_SCRAMBLED=2
+} scrambled_t;
+
 typedef struct {
     int video_attr;
+    scrambled_t scrambled;
 } p_program_attr_t;
 p_program_attr_t* ifo_program_attrs;
 
@@ -460,7 +501,7 @@ typedef struct {
         uint32_t info_260_sa;    /* ? start address */
         uint8_t  zero_264[3];
         struct {
-            uint8_t enabled;     /* Encrypted Title Key Status */
+            uint8_t supported;   /* Encrypted Title Key Status */
             uint8_t title_key[8];/* This needs to be decrypted using media key */
         } cprm;
         uint8_t  zero_276[28];
@@ -470,7 +511,7 @@ typedef struct {
         uint32_t info_312_sa;    /* user defined program set info start address? */
         uint32_t info_316_sa;    /* ? start address */
         uint8_t  zero_320[32];
-        uint32_t info_352_sa;    /* ? start address */
+        uint32_t txt_attr_sa;    /* extra attributes for programs (chan id etc.) */
         uint32_t info_356_sa;    /* ? start address */
         uint8_t  zero_360[152];
     } PACKED mat;
@@ -548,7 +589,7 @@ typedef struct  {
     char     title[64];       /* Could be same as label, NUL, or another charset */
     uint16_t prog_set_id;     /* On LG V1.1 discs this is program set ID */
     uint16_t first_prog_id;   /* ID of first program in this program set */
-    char     data4[6];
+    char     data3[6];
 } PACKED psi_t;
 
 static const char* parse_txt_encoding(uint8_t txt_encoding)
@@ -717,6 +758,34 @@ static bool parse_pgtm(pgtm_t pgtm, struct tm* tm)
     return ret;
 }
 
+#ifndef NDEBUG
+/* This is basically a simplification of find_program_text_info() */
+static void print_psi(psi_gi_t* psi_gi)
+{
+    putc('\n', stdinfo);
+    int ps;
+    uint16_t program_count = 0;
+    for (ps=0; ps<psi_gi->nr_of_psi; ps++) {
+        psi_t *psi = (psi_t*)(((char*)(psi_gi+1)) + (ps * sizeof(psi_t)));
+
+        uint16_t first_prog_num = ntohs(psi->first_prog_id); /* assuming this is first to play? */
+        uint16_t start_prog_num = program_count+1;
+        uint16_t num_progs_in_set = ntohs(psi->nr_of_programs);
+        program_count += num_progs_in_set;
+        fprintf(stdinfo, "Programs in Program set %d:", ps+1);
+        int program_id;
+        for (program_id = start_prog_num;
+             program_id < start_prog_num+num_progs_in_set;
+             program_id++) {
+            const char* fmt = (program_id==first_prog_num ? " (%d)" : " %d");
+            fprintf(stdinfo, fmt, program_id);
+        }
+        putc('\n', stdinfo);
+    }
+    putc('\n', stdinfo);
+}
+#endif//NDEBUG
+
 /*
  * FIXME: This assumes the programs occur linearly within
  * the default program sets. This has been accurate for all
@@ -729,14 +798,24 @@ static psi_t* find_program_text_info(psi_gi_t* psi_gi, int program)
     uint16_t program_count = 0;
     for (ps=0; ps<psi_gi->nr_of_psi; ps++) {
         psi_t *psi = (psi_t*)(((char*)(psi_gi+1)) + (ps * sizeof(psi_t)));
-        uint16_t start_prog_num = ntohs(psi->first_prog_id);
-        if (start_prog_num==0 || start_prog_num==0xFFFF) {
-            /* Need to maintain program count if first_prog_id not stored,
-             * as is the case for LG and "CIRRUS LOGIC" V1.1 discs for example. */
-            start_prog_num = program_count+1;
-            program_count += ntohs(psi->nr_of_programs);
-        }
-        uint16_t end_prog_num = start_prog_num + ntohs(psi->nr_of_programs) - 1;
+        uint16_t start_prog_num;
+        /*
+        start_prog_num = ntohs(psi->first_prog_id);
+
+        We need to maintain program count as first_prog_id is often not stored,
+        as is the case for LG and "CIRRUS LOGIC" V1.1 discs for example (it's 0 or 0xFFFF).
+        Also I noticed a Sony disc that had programs sets with 2 programs in them, which
+        sometimes set the first_prog_id to the second program in the set.  Perhaps
+        this field identifies the prog to start playing, as the first program in those
+        sets was a single VOBU that was generated due to a split.
+        Perhaps I should name the programs label.ps_id(001) when > 1 ps. */
+        start_prog_num = program_count+1;
+        uint16_t num_progs_in_set = ntohs(psi->nr_of_programs);
+        /* TODO: Perhaps have an option to merge all programs
+           in a program set to a vob using this info. That would assume
+           though that the programs were adjacent. */
+        program_count += num_progs_in_set;
+        uint16_t end_prog_num = start_prog_num + num_progs_in_set - 1;
         if ((program >= start_prog_num) && (program <= end_prog_num)) {
             return psi;
         }
@@ -753,8 +832,7 @@ static psi_t* find_program_text_info(psi_gi_t* psi_gi, int program)
  */
 static char* text_field_convert(const char* field, unsigned int len)
 {
-    /* UTF-8 can have up to 6 bytes per char + space for \0 */
-    unsigned int conv_max_len=len*6+1;
+    unsigned int conv_max_len=len*MB_LEN_MAX+1/*NUL*/;
     char* field_local=malloc(conv_max_len);
     if (!field_local) {
         fprintf(stderr, "Error allocating space for text conversion\n");
@@ -820,6 +898,89 @@ static void print_disc_info(rtav_vmgi_t* rtav_vmgi_ptr)
     }
 }
 
+static char* mb_clean_name(const char* src)
+{
+    size_t src_size = strlen (src) + 1;
+    wchar_t *str_wc = NULL;
+    size_t src_chars = mbstowcs (NULL, src, 0);
+    if (src_chars == (size_t) -1)
+        return NULL;
+    src_chars += 1; /* make space for NUL */
+    str_wc = malloc (src_chars * sizeof (wchar_t));
+    if (str_wc == NULL)
+        return NULL;
+    if (mbstowcs (str_wc, src, src_chars) <= 0) {
+        free(str_wc);
+        return NULL;
+    }
+    str_wc[src_chars - 1] = L'\0';
+
+    wchar_t* wc = str_wc;
+    while (*wc) {
+        size_t good = wcscspn(wc, L" /:?\\");
+        if (good)
+            wc+=good;
+        else
+            *wc++=L'-';
+    }
+
+    char* newstr = malloc (src_size);
+    if (newstr == NULL) {
+        free(str_wc);
+        return NULL;
+    }
+    (void) wcstombs(newstr, str_wc, src_size);
+    free(str_wc);
+
+    return newstr;
+}
+
+/* Must pass a string on the heap which may be modified inplace,
+ * or may be reallocated. */
+static char* clean_name(char* src, bool mb_src)
+{
+    if (mb_src && (MB_CUR_MAX > 1)) {
+        char* cleaned = mb_clean_name(src);
+        free(src);
+        return cleaned;
+    }
+
+    char* c = src;
+    while (*c) {
+        size_t good = strcspn(c, " /:?\\");
+        if (good)
+            c+=good;
+        else
+            *c++='-';
+    }
+    return src;
+}
+
+static char* get_label_base(const psi_t* psi)
+{
+    char* title_local = text_field_convert(psi->title, sizeof(psi->title));
+    if (title_local && *title_local &&
+        strncmp(title_local, psi->label, sizeof(psi->label))) { /* if title != label */
+        title_local = clean_name(title_local, true);
+        if (!title_local) {
+          fprintf(stderr, "Error generating file name from title\n");
+          return NULL;
+        } else {
+          return title_local;
+        }
+    }
+    free(title_local);
+
+    const char* label=psi->label; /* ASCII */
+    if (*label && !STREQ(label, " ")) {
+        char* label_local = my_strndup(label, sizeof(psi->label));
+        label_local = clean_name(label_local, false);
+        return label_local;
+    }
+
+    return NULL;
+}
+
 static void print_label(const psi_t* psi)
 {
     const char* label=psi->label; /* ASCII */
@@ -860,11 +1021,13 @@ Specifically for widescreen PAL movies my DVD-VR camcorder set:
 #define MPEG_HEADER_LEN 4
 #define SEQUENCE_ID 0xB3
 #define SEQUENCE_EXTENSION_ID 0xB5
+#define VIDEO_STREAM_0 0xE0 /* I've only seen E0 on dvd-vr discs (E0-F possible) */
 #define SEQUENCE_LEN 4 /* length of data we need to parse from sequence packet */
 #define SEQUENCE_EXTENSION_LEN 5 /* length of data we need to parse from sequence extension packet */
+#define VIDEO_STREAM_LEN 3 /* length of data we need to parse from video stream packet */
 
 /* Return offset to header or -1 if not found */
-static int find_mpeg2_header(const uint8_t* buf, const unsigned int bs, const uint8_t type)
+static int find_mpeg_header(const uint8_t* buf, const unsigned int bs, const uint8_t type)
 {
     unsigned int offset=0;
     uint32_t header = 0x00000100 + type;
@@ -910,7 +1073,7 @@ static p_video_attr_t get_sequence_display_extension_sizes(const uint8_t* buf, c
 
     uint8_t type = *(buf + offset + MPEG_HEADER_LEN);
     int skip_colour=(type&0x01) ? 3 : 0;
-    const uint8_t* display_size = buf + offset + MPEG_HEADER_LEN + skip_colour + 1;
+    const uint8_t* display_size = buf + offset + MPEG_HEADER_LEN + skip_colour + sizeof (type);
     uint16_t horiz_disp_size  = *(display_size) << 6;
             horiz_disp_size += *(display_size+1) >> 2;
     uint16_t vert_disp_size = (*(display_size+1) & 0x01) << 13;
@@ -930,7 +1093,7 @@ static void set_sequence_display_extension_sizes(uint8_t* buf, const unsigned in
 
     uint8_t type = *(buf + offset + MPEG_HEADER_LEN);
     int skip_colour=(type&0x01) ? 3 : 0;
-    uint8_t* display_size = buf + offset + MPEG_HEADER_LEN + skip_colour + 1;
+    uint8_t* display_size = buf + offset + MPEG_HEADER_LEN + skip_colour + sizeof (type);
 
     /* One could precalc this per program, but frequency is low so not worth the effort */
     *(uint32_t*)display_size = htonl(0x00020000);
@@ -941,6 +1104,32 @@ static void set_sequence_display_extension_sizes(uint8_t* buf, const unsigned in
     *(display_size+3) |= (vert_disp_size << 3);
 }
 
+static void check_mpeg_encryption(uint8_t* buf, const unsigned int bs, const unsigned int program)
+{
+    /* Note we'll warn below if we've not seen any video stream (E0 doesn't match).
+     * Note also I've only seen AC-3 audio on 0xBD and it has also been
+     * encrypted on discs I've seen.  */
+    if (ifo_program_attrs[program].scrambled != PARTIALLY_SCRAMBLED) {
+        int pes_offset = find_mpeg_header(buf, bs-VIDEO_STREAM_LEN, VIDEO_STREAM_0);
+        if (pes_offset >= 0) {
+            /* extension header is always available for 0xBD and 0xE? types */
+            uint8_t scramble_byte = *(buf + pes_offset + MPEG_HEADER_LEN + 2);
+            bool scrambled;
+            if ((scramble_byte & 0xC0) == 0x80) { /* MPEG2 */
+                scrambled = scramble_byte & 0x30;
+            } else {
+                scrambled = false; /* assuming MPEG1 doesn't support encryption */
+            }
+            if (ifo_program_attrs[program].scrambled != SCRAMBLED_UNSET &&
+                ifo_program_attrs[program].scrambled != scrambled) {
+                ifo_program_attrs[program].scrambled = PARTIALLY_SCRAMBLED;
+            } else {
+                ifo_program_attrs[program].scrambled = scrambled;
+            }
+        }
+    }
+}
+
 static void fix_mpeg2_aspect(uint8_t* buf, const unsigned int bs, const unsigned int program)
 {
     static int sector;
@@ -956,7 +1145,7 @@ static void fix_mpeg2_aspect(uint8_t* buf, const unsigned int bs, const unsigned
     p_video_attr_t s_video_attr = { .aspect=ifo_video_attr.aspect, .width=-1, .height=-1 };
 
     if (sequence_offset == -1) {
-        if ((sequence_offset = find_mpeg2_header(buf, bs-SEQUENCE_LEN, SEQUENCE_ID)) >= 0) {
+        if ((sequence_offset = find_mpeg_header(buf, bs-SEQUENCE_LEN, SEQUENCE_ID)) >= 0) {
             found_sequence_header = true;
 #ifndef NDEBUG
             fprintf(stdinfo,"Found SH  @ %d+%d\n", sector, sequence_offset);
@@ -966,19 +1155,20 @@ static void fix_mpeg2_aspect(uint8_t* buf, const unsigned int bs, const unsigned
                 set_sequence_aspect(buf, sequence_offset, s_video_attr);
             }
         }
-    } else if (sequence_aspect!=ifo_video_attr.aspect) {
-        if (find_mpeg2_header(buf+sequence_offset, MPEG_HEADER_LEN, SEQUENCE_ID)==0) {
+    } else {
+        if (find_mpeg_header(buf+sequence_offset, MPEG_HEADER_LEN, SEQUENCE_ID)==0) {
             found_sequence_header = true;
 #ifndef NDEBUG
             fprintf(stdinfo,"Found SH  @ %d+%d\n", sector, sequence_offset);
 #endif
-            set_sequence_aspect(buf, sequence_offset, s_video_attr);
+            if (sequence_aspect!=ifo_video_attr.aspect)
+                set_sequence_aspect(buf, sequence_offset, s_video_attr);
         } else if (look_harder) {
         /* I can't see why the sequence headers would be at arbitrary offsets in each sector,
          * and I've analyzed about 10 different VROs and they all have the same offsets.
          * So I think that doing this will only redundantly scan every byte of sectors
          * without sequence headers. */
-            int curr_offset = find_mpeg2_header(buf, bs-SEQUENCE_LEN, SEQUENCE_ID);
+            int curr_offset = find_mpeg_header(buf, bs-SEQUENCE_LEN, SEQUENCE_ID);
             if (curr_offset >= 0) {
                 found_sequence_header = true;
                 sequence_offset = curr_offset;
@@ -1000,9 +1190,9 @@ static void fix_mpeg2_aspect(uint8_t* buf, const unsigned int bs, const unsigned
         }
         int extension_offset = look_harder ? 0 : sequence_offset + MPEG_HEADER_LEN + SEQUENCE_LEN;
         int next_offset;
-        while ((next_offset = find_mpeg2_header(buf + extension_offset,
-                                                bs - extension_offset - SEQUENCE_EXTENSION_LEN,
-                                                SEQUENCE_EXTENSION_ID)) >= 0) {
+        while ((next_offset = find_mpeg_header(buf + extension_offset,
+                                               bs - extension_offset - SEQUENCE_EXTENSION_LEN,
+                                               SEQUENCE_EXTENSION_ID)) >= 0) {
             extension_offset += next_offset;
             uint8_t type = *(buf + extension_offset + MPEG_HEADER_LEN);
             if ((type&0xF0) == 0x20) {
@@ -1033,7 +1223,7 @@ static void fix_mpeg2_aspect(uint8_t* buf, const unsigned int bs, const unsigned
 /* Haven't had a request to do this yet:
    http://forum.doom9.org/archive/index.php/t-102969.html
    Note code there makes incorrect assumptions about offsets I think.  */
-static void add_mpeg2_nav(uint8_t* buf, const unsigned int bs)
+static void add_mpeg_nav(uint8_t* buf, const unsigned int bs)
 {
     (void) buf; (void) bs;
 }
@@ -1041,7 +1231,8 @@ static void add_mpeg2_nav(uint8_t* buf, const unsigned int bs)
 void process_mpeg2(uint8_t* buf, const unsigned int bs, void* program)
 {
     fix_mpeg2_aspect(buf, bs, *(const unsigned int*)program);
-    add_mpeg2_nav(buf, bs);
+    add_mpeg_nav(buf, bs);
+    check_mpeg_encryption(buf, bs, *(const unsigned int*)program);
 }
 
 /*********************************************************************************
@@ -1067,6 +1258,8 @@ static void usage(char** argv, int error)
                    "  -n, --name=NAME    Specify a basename to use for extracted vob files\n"
                    "                     rather than using one based on the timestamp.\n"
                    "                     If you pass `-' the vob files will be written to stdout.\n"
+                   "                     If you pass `[label]' the names will be based on\n"
+                   "                     a sanitized version of the title or label.\n"
                    "\n"
                    "      --help         Display this help and exit.\n"
                    "      --version      Output version information and exit.\n"
@@ -1171,7 +1364,7 @@ int main(int argc, char** argv)
     }
 
     int vro_fd=-1;
-    if (vro_name && !rtav_vmgi_ptr->mat.cprm.enabled) {
+    if (vro_name) {
         vro_fd=open(vro_name,O_RDONLY);
         if (vro_fd == -1) {
             fprintf(stderr, "Error opening [%s] (%s)\n", vro_name, strerror(errno));
@@ -1186,8 +1379,10 @@ int main(int argc, char** argv)
     rtav_vmgi_ptr->mat.version &= 0x00FF;
     fprintf(stdinfo, "format: DVD-VR V%d.%d\n",
             rtav_vmgi_ptr->mat.version>>4,rtav_vmgi_ptr->mat.version&0x0F);
-    if (rtav_vmgi_ptr->mat.cprm.enabled) {
-        fprintf(stdinfo, "Encryption: CPRM\n");
+    if (rtav_vmgi_ptr->mat.cprm.supported) {
+        fprintf(stdinfo, "Encryption: CPRM supported\n");
+        /* Note programs may not actually be encrypted.
+         * That's indicated per AV pack in 2 PES scrambling control bits */
     }
 
     disc_charset=parse_txt_encoding(rtav_vmgi_ptr->mat.txt_encoding);
@@ -1202,6 +1397,10 @@ int main(int argc, char** argv)
     psi_gi_t *def_psi_gi = (psi_gi_t*) ((char*)rtav_vmgi_ptr + rtav_vmgi_ptr->mat.def_psi_sa);
 
 #ifndef NDEBUG
+    NTOHS(def_psi_gi->nr_of_programs);
+    if ((def_psi_gi->nr_of_psi > 1) && (def_psi_gi->nr_of_psi != def_psi_gi->nr_of_programs)) {
+        print_psi(def_psi_gi);
+    }
     fprintf(stdinfo, "Number of info tables for VRO: %d\n",pgiti->nr_of_pgi);
     fprintf(stdinfo, "Number of vob formats: %d\n",pgiti->nr_of_vob_formats);
     fprintf(stdinfo, "pgit_ea: %08"PRIX32"\n",pgiti->pgit_ea);
@@ -1283,7 +1482,7 @@ int main(int argc, char** argv)
         vvob_t* vvob = (vvob_t*) (((uint8_t*)pgiti) + *vvobi_sa);
         struct tm tm;
         bool ts_ok = parse_pgtm(vvob->vob_timestamp,&tm);
-        char vob_base[32];
+        char vob_base[(sizeof(psi->title)*MB_LEN_MAX)+4/*#123*/+1/*NUL*/];
         if (STREQ(base_name, TIMESTAMP_FMT)) { //use timestamp to give unique filename
             if (ts_ok) {
                 strftime(vob_base,sizeof(vob_base),TIMESTAMP_FMT,&tm);
@@ -1292,6 +1491,22 @@ int main(int argc, char** argv)
                 int datelen=strlen(vob_base);
                 (void) snprintf(vob_base+datelen, sizeof(vob_base)-datelen, "#%03d", program+1);
             }
+        } else if (STREQ(base_name, "[label]")) { //use the label to generate filename
+            if (!psi) {
+                fprintf(stderr, "Error: Couldn't generate name based on label\n");
+                exit(EXIT_FAILURE);
+            }
+            char* label_base = get_label_base(psi);
+            if (!label_base) {
+                fprintf(stderr, "Error: Couldn't generate name based on empty label\n");
+                exit(EXIT_FAILURE);
+            }
+            unsigned int ret = snprintf(vob_base, sizeof(vob_base), "%s#%03d", label_base, program+1);
+            if (ret >= sizeof(vob_base)) { /* Shouldn't happen.  */
+                fprintf(stderr, "Error: label is too long\n");
+                exit(EXIT_FAILURE);
+            }
+            free(label_base);
         } else {
             unsigned int ret = snprintf(vob_base, sizeof(vob_base), "%s#%03d", base_name, program+1);
             if (ret >= sizeof(vob_base)) {
@@ -1301,12 +1516,12 @@ int main(int argc, char** argv)
         }
 
         int vob_fd=-1;
-        char vob_name[64];
+        char vob_name[sizeof(vob_base)+4];
         if (vro_fd!=-1) {
             if (STREQ(base_name, "-")) {
                 vob_fd=fileno(stdout);
             } else {
-                (void) snprintf(vob_name,sizeof(vob_name),"%s.vob",vob_base); /* 1 char too long for ls -l in 80 cols :( */
+                (void) snprintf(vob_name,sizeof(vob_name),"%s.vob",vob_base);
                 vob_fd=open(vob_name,O_WRONLY|O_CREAT|O_EXCL,0666);
                 if (vob_fd == -1 && errno == EEXIST && STREQ(base_name, TIMESTAMP_FMT)) {
                     /* JVC DVD recorder can generate duplicate timestamps at least :( */
@@ -1327,6 +1542,7 @@ int main(int argc, char** argv)
             fprintf(stdinfo, "vob format: %d\n", vvob->vob_format_id);
         }
         ifo_program_attrs[program].video_attr = vvob->vob_format_id-1;
+        ifo_program_attrs[program].scrambled = SCRAMBLED_UNSET;
 
         NTOHS(vvob->vob_attr);
         int skip=0;
@@ -1365,6 +1581,7 @@ int main(int argc, char** argv)
         int vobus;
         uint64_t tot=0;
         int display_char;
+        bool processed_some_video = false;
         int error=0;
         if (vro_fd != -1) {
             percent_display(PERCENT_START, 0, 0);
@@ -1401,8 +1618,13 @@ int main(int argc, char** argv)
                             exit(EXIT_FAILURE);
                         }
                     }
+                } else if (ifo_program_attrs[program].scrambled == SCRAMBLED ||
+                           ifo_program_attrs[program].scrambled == PARTIALLY_SCRAMBLED) {
+                    display_char='E';
+                    processed_some_video = true;
                 } else {
                     display_char=0; /* default */
+                    processed_some_video = true;
                 }
 
                 int percent=((vobus+1)*100)/vobu_map->nr_of_vobu_info;
@@ -1426,11 +1648,16 @@ int main(int argc, char** argv)
 
         fprintf(stdinfo, "size : %'"PRIu64"\n",tot*DVD_SECTOR_SIZE);
 
-        vvobi_sa++;
-    }
+        if (ifo_program_attrs[program].scrambled == SCRAMBLED) {
+            fprintf(stderr, "Warning: program is encrypted\n");
+        } else if (ifo_program_attrs[program].scrambled == PARTIALLY_SCRAMBLED) {
+            fprintf(stderr, "Warning: program is partially encrypted\n");
+        } else if (ifo_program_attrs[program].scrambled == SCRAMBLED_UNSET && processed_some_video) {
+            fprintf(stderr, "Warning: didn't detect a video stream, please report\n");
+            fprintf(stderr, "  (preferably with a sample vob file)\n");
+        }
 
-    if (vro_name && rtav_vmgi_ptr->mat.cprm.enabled) {
-        fprintf(stderr, "\nError: This disc employs CPRM encryption which is currently unsupported.\n");
+        vvobi_sa++;
     }
 
     free(ifo_program_attrs);
